from bifabrik.dst.DataDestination import DataDestination
from bifabrik.base.Pipeline import Pipeline
from bifabrik.cfg.TableDestinationConfiguration import TableDestinationConfiguration
from pyspark.sql.dataframe import DataFrame
from bifabrik.utils import fsUtils

class TableDestination(DataDestination, TableDestinationConfiguration):
    """Saves data to a lakehouse table.

    Examples
    --------
    > import bifabrik as bif
    >
    > bif.fromSql.query('SELECT * FROM SomeTable').toTable('DestinationTable1').run()
    """
    def __init__(self, pipeline: Pipeline, targetTableName: str):
        super().__init__(pipeline)
        TableDestinationConfiguration.__init__(self)
        self._targetTableName = targetTableName

    def __str__(self):
        return f'Table destination: {self._targetTableName}'
    
    def execute(self, input: DataFrame) -> None:
        mergedConfig = self._pipeline.configuration.mergeToCopy(self)
        dstLh = mergedConfig.destinationStorage.destinationLakehouse
        dstWs = mergedConfig.destinationStorage.destinationWorkspace
        basePath = fsUtils.getLakehousePath(dstLh, dstWs)

        input.write.mode("overwrite").format("delta").option("overwriteSchema", "true").save(basePath + "/Tables/" + self._targetTableName)


import json
from bifabrik.cfg.engine.ConfigContainer import ConfigContainer

from bifabrik.cfg.specific.LogConfiguration import LogConfiguration
from bifabrik.cfg.specific.CsvConfiguration import CsvConfiguration
from bifabrik.cfg.specific.JsonConfiguration import JsonConfiguration
from bifabrik.cfg.specific.FileSourceConfiguration import FileSourceConfiguration
from bifabrik.cfg.specific.SourceStorageConfiguration import SourceStorageConfiguration
from bifabrik.cfg.specific.DestinationStorageConfiguration import DestinationStorageConfiguration

import notebookutils.mssparkutils.fs

class CompleteConfiguration(ConfigContainer):
    """
    Fluent API for configuration
    
    You can set configuration directly through setter methods like .delimiter(';')[.further options...]
    or
    .option(name, value = None) gets / sets a named config property as in
    .option('delimiter', ';') #setter
    .option('delimiter') #getter
    """
    def __init__(self):
        self.__log = LogConfiguration()
        self.__csv = CsvConfiguration()
        self.__json = JsonConfiguration()
        self.__fileSource = FileSourceConfiguration()
        self.__sourceStorage = SourceStorageConfiguration()
        self.__destinationStorage = DestinationStorageConfiguration()
        super().__init__()
        

    @property
    def csv(self) -> CsvConfiguration:
        """CSV files settings"""
        return self.__csv
    
    @property
    def json(self) -> JsonConfiguration:
        """JSON files settings"""
        return self.__json
    
    @property
    def log(self) -> LogConfiguration:
        """Logging settings"""
        return self.__log
    
    @property
    def fileSource(self) -> FileSourceConfiguration:
        """General file loading settings"""
        return self.__fileSource
    
    @property
    def sourceStorage(self) -> SourceStorageConfiguration:
        """The source lakehouse and warehouse reference used by pipelines"""
        return self.__sourceStorage

    @property
    def destinationStorage(self) -> DestinationStorageConfiguration:
        """The destination lakehouse and warehouse reference used by pipelines"""
        return self.__destinationStorage
    
    def serialize(self) -> str:
        """Serializes the configuration to a string"""
        res = {}
        rootAttribs = dir(self)
        noUnderscoreRootAttrNames = list(filter(lambda x: not x.startswith("_"), rootAttribs))
        for rootAttrName in noUnderscoreRootAttrNames:
            rootAttr = getattr(self, rootAttrName)
            attrType = type(rootAttr)
            if attrType.__bases__[0].__name__ == 'Configuration':
                res[rootAttrName] = rootAttr._explicitProps
        j = json.dumps(res, indent=4)
        return j

    @staticmethod
    def deserialize(json_data):
        """Deserializes a CompleteConfiguration from a string"""
        json_dict = json.loads(json_data)
        cfg = CompleteConfiguration()
        for cfgAttrName in json_dict:
            for cfgPropName in json_dict[cfgAttrName]:
                cfg.option(cfgPropName, json_dict[cfgAttrName][cfgPropName])
        return cfg
    
    def saveToFile(self, path):
        """Saves the configuration to a JSON file"""
        serialized = self.serialize()
        notebookutils.mssparkutils.fs.put(path, serialized, True)

    def loadFromFile(self, path):
        """Loads the configuration values from a JSON file"""
        json_data = notebookutils.mssparkutils.fs.head(path, 1024 * 1024 * 1024)
        deser = CompleteConfiguration.deserialize(json_data)
        self.merge(deser)
        return self
        


from bifabrik.cfg.engine.Configuration import Configuration
from bifabrik.cfg.engine.Configuration import CfgProperty

class FileSourceConfiguration(Configuration):
    """Configuration related to loading data from files (aspects common for all file formats)
    """
    def __init__(self):
        self._explicitProps = {}
        self.__encoding = None
        self.__locale = None

    @CfgProperty
    def encoding(self) -> str:
        """From pandas:
        How encoding errors are treated. https://docs.python.org/3/library/codecs.html#standard-encodings
        """
        return self.__encoding
    @encoding.setter(key='encoding')
    def encoding(self, val):
        self.__encoding = val

    @CfgProperty
    def locale(self) -> str:
        """From spark:
        sets a locale as language tag in IETF BCP 47 format. If None is set, it uses the default value, en-US. For instance, locale is used while parsing dates and timestamps.
        """
        return self.__locale
    @locale.setter(key='locale')
    def jsonLocale(self, val):
        self.__locale = val


from bifabrik.src.DataSource import DataSource
from bifabrik.cfg.CsvSourceConfiguration import CsvSourceConfiguration
from pyspark.sql.dataframe import DataFrame
import pandas as pd
from bifabrik.utils import fsUtils
from bifabrik.utils import log

class CsvSource(DataSource, CsvSourceConfiguration):
    """CSV data source
    uses the pandas loaders
    supports configuration options from https://pandas.pydata.org/docs/reference/api/pandas.read_csv.html
    
    Examples
    --------
    > import bifabrik as bif
    >
    > bif.fromCsv.path('DATA/factOrderLine*.csv').delimiter(';').decimal(',').toTable('FactOrderLine').run()
    """
    
    def __init__(self, parentPipeline):
        super().__init__(parentPipeline)
        CsvSourceConfiguration.__init__(self)
        self._path = ""

    def __str__(self):
        return f'CSV source: {self._path}'
    
    def path(self, path: str):
        """Set the path (or pattern) to the source file.
        It searches the Files/ folder in the current lakehouse
        """
        self._path = path
        return self
    
    def execute(self, input):
        lgr = log.getLogger()
        mergedConfig = self._pipeline.configuration.mergeToCopy(self)

        srcLh = mergedConfig.sourceStorage.sourceLakehouse
        srcWs = mergedConfig.sourceStorage.sourceWorkspace

        # pandas seems to have a problem with absolute ABFS paths, so use /lakehouse/default/ if refering to the default lakehouse
        source_files = fsUtils.filePatternSearch(self._path, srcLh, srcWs, useImplicitDefaultLakehousePath = True)
        if len(source_files) == 0:
            return None
        
        separator = ', '
        lgr.info(f'Loading CSV files: [{separator.join(source_files)}]')
        
        #source_files = fsUtils.filePatternSearch(self._path)
        fileDfs = []
        for src_file in source_files:
            #csv_path = f'/lakehouse/default/{src_file}'
            pd_df = pd.read_csv(
                filepath_or_buffer = src_file, 
                encoding = mergedConfig.fileSource.encoding,
                delimiter = mergedConfig.csv.delimiter,
                header = mergedConfig.csv.header,
                thousands = mergedConfig.csv.thousands,
                decimal = mergedConfig.csv.decimal,
                quotechar = mergedConfig.csv.quotechar,
                quoting = mergedConfig.csv.quoting,
                escapechar = mergedConfig.csv.escapechar
                )
            df = self._spark.createDataFrame(pd_df)
            fileDfs.append(df)
        if len(fileDfs) == 0:
            return None
        df = fileDfs[0]
        if len(fileDfs) > 1:
            for i in range(1, len(fileDfs)):
                df = df.union(fileDfs[i])
        
        self._result = df
        self._completed = True


from bifabrik.cfg.engine.ConfigContainer import ConfigContainer
from bifabrik.cfg.specific.CsvConfiguration import CsvConfiguration
from bifabrik.cfg.specific.FileSourceConfiguration import FileSourceConfiguration
from bifabrik.cfg.DataSourceConfigurationBase import DataSourceConfigurationBase

#class CsvSourceConfiguration(ConfigContainer):
class CsvSourceConfiguration(DataSourceConfigurationBase):

    def __init__(self):
        self.__csv = CsvConfiguration()
        self.__fileSource = CsvConfiguration()
        super().__init__()
        

    @property
    def csv(self) -> CsvConfiguration:
        return self.__csv
    
    @property
    def fileSource(self) -> FileSourceConfiguration:
        return self.__fileSource


from bifabrik.cfg.engine.ConfigContainer import ConfigContainer
from bifabrik.cfg.DataDestinationConfigurationBase import DataDestinationConfigurationBase

class TableDestinationConfiguration(DataDestinationConfigurationBase):

    def __init__(self):
        super().__init__()


from bifabrik.cfg.engine.Configuration import Configuration
from bifabrik.cfg.engine.Configuration import CfgProperty

class SourceStorageConfiguration(Configuration):
    """Defines the data source lakehouse. By default refers to the default lakehouse of the notebook.
    """
    def __init__(self):
        self._explicitProps = {}
        self.__sourceWorkspace = None
        self.__sourceLakehouse = None

    @CfgProperty
    def sourceWorkspace(self) -> str:
        """The workspace from which pipeline sources load data - ID or name
        """
        return self.__sourceWorkspace
    @sourceWorkspace.setter(key='sourceWorkspace')
    def sourceWorkspace(self, val):
        self.__sourceWorkspace = val

    @CfgProperty
    def sourceLakehouse(self) -> str:
        """The lakehouse from which pipeline sources load data - ID or name
        """
        return self.__sourceLakehouse
    @sourceLakehouse.setter(key='sourceLakehouse')
    def sourceLakehouse(self, val):
        self.__sourceLakehouse = val


from bifabrik.cfg.engine.ConfigContainer import ConfigContainer
#from bifabrik.base.Pipeline import Pipeline

class Task:
    
    # parentPipeline: Pipeline
    def __init__(self, parentPipeline):
        parentPipeline.addTask(self)
        self._pipeline = parentPipeline
        self._spark = parentPipeline.spark
        self._result = None
        self._error = None
        self._completed = False

    def execute(self, input):
        """Obtain the result (typically a dataframe) or save the data in case of a destination
        """
        pass

    def cleanup(self):
        """Cleanup after the execution (archive processed files, remove temp data)
        """
        pass

    def rollback(self):
        """Revert changes made during a failed execution (restore modified delta tables or processed files)
        """
        pass

    @property
    def error(self):
        """The error message from the last execution (str)
        """
        return self._error
    
    @property
    def result(self):
        """The result (if any) of a completed task - typically a pyspark.sql.dataframe.DataFrame"""
        return self._result
    
    @property
    def completed(self):
        """Indicator of the task being completed"""
        return self._completed
    
    def getTaskResult(self) -> any:
        """Execute all tasks in the pipeline up to this one and return its result
        """
        return self._pipeline.getTaskResult(self)

    def clearResults(self):
        self._result = None
        self._error = None
        self._completed = False



class Configuration:
    pass

class CfgProperty(property):
    def __set__(self, obj, value):
        super().__set__(obj, value)
        att = getattr(obj, '_explicitProps')
        att[self.key] = value

    def _setter(self, fset):
        obj = super().setter(fset)
        obj.key = self.key
        return obj

    def setter(self, key):
        self.key = key
        return self._setter


from bifabrik.cfg.engine.Configuration import Configuration
from bifabrik.cfg.engine.Configuration import CfgProperty

class CsvConfiguration(Configuration):
    """Configuration related to loading CSV files
    """
    def __init__(self):
        self._explicitProps = {}
        self.__delimiter = ','
        self.__header = 'infer'
        self.__thousands = None
        self.__decimal = '.'
        self.__quotechar = '"'
        self.__quoting = 0
        self.__escapechar = None

    @CfgProperty
    def delimiter(self) -> str:
        """From pandas:
        Character or regex pattern to treat as the delimiter. If sep=None, the C engine cannot automatically detect the separator, but the Python parsing engine can, meaning the latter will be used and automatically detect the separator from only the first valid row of the file by Pythons builtin sniffer tool
        """
        return self.__delimiter
    @delimiter.setter(key='delimiter')
    def delimiter(self, val):
        self.__delimiter = val

    @CfgProperty
    def header(self) -> str:
        """From pandas:
        Row number(s) containing column labels and marking the start of the data (zero-indexed). Default behavior is to infer the column names: if no names are passed the behavior is identical to header=0 and column names are inferred from the first line of the file, if column names are passed explicitly to names then the behavior is identical to header=None. Explicitly pass header=0 to be able to replace existing names. The header can be a list of integers that specify row locations for a MultiIndex on the columns e.g. [0, 1, 3]. Intervening rows that are not specified will be skipped (e.g. 2 in this example is skipped). Note that this parameter ignores commented lines and empty lines if skip_blank_lines=True, so header=0 denotes the first line of data rather than the first line of the file.
        """
        return self.__header
    @header.setter(key='header')
    def header(self, val):
        self.__header = val

    @CfgProperty
    def thousands(self) -> str:
        """From pandas:
        Character acting as the thousands separator in numerical values.
        """
        return self.__thousands
    @thousands.setter(key='thousands')
    def thousands(self, val):
        self.__thousands = val

    @CfgProperty
    def decimal(self) -> str:
        """From pandas:
        Character to recognize as decimal point (e.g., use ‘,’ for European data).
        """
        return self.__decimal
    @decimal.setter(key='decimal')
    def decimal(self, val):
        self.__decimal = val

    @CfgProperty
    def quotechar(self) -> str:
        """From pandas:
        Character used to denote the start and end of a quoted item. Quoted items can include the delimiter and it will be ignored.
        """
        return self.__quotechar
    @quotechar.setter(key='quotechar')
    def quotechar(self, val):
        self.__quotechar = val

    @CfgProperty
    def quoting(self) -> int:
        """From pandas:
        {0 or csv.QUOTE_MINIMAL, 1 or csv.QUOTE_ALL, 2 or csv.QUOTE_NONNUMERIC, 3 or csv.QUOTE_NONE}, default csv.QUOTE_MINIMAL
        Control field quoting behavior per csv.QUOTE_* constants. Default is csv.QUOTE_MINIMAL (i.e., 0) which implies that only fields containing special characters are quoted (e.g., characters defined in quotechar, delimiter, or lineterminator.
        """
        return self.__quoting
    @quoting.setter(key='quoting')
    def quoting(self, val):
        self.__quoting = val

    @CfgProperty
    def escapechar(self) -> int:
        """From pandas:
        Character used to escape other characters.
        """
        return self.__escapechar
    @escapechar.setter(key='escapechar')
    def escapechar(self, val):
        self.__escapechar = val


from bifabrik.cfg.engine.Configuration import Configuration
from bifabrik.cfg.engine.Configuration import CfgProperty

class DestinationStorageConfiguration(Configuration):
    """Defines the data destination lakehouse. By default refers to the default lakehouse of the notebook.
    """
    def __init__(self):
        self._explicitProps = {}
        self.__destinationWorkspace = None
        self.__destinationLakehouse = None

    @CfgProperty
    def destinationWorkspace(self) -> str:
        """The workspace to which pipeline destinations writes data - ID or name
        """
        return self.__destinationWorkspace
    @destinationWorkspace.setter(key='destinationWorkspace')
    def destinationWorkspace(self, val):
        self.__destinationWorkspace = val

    @CfgProperty
    def destinationLakehouse(self) -> str:
        """The lakehouse to which pipeline destinations writes data - ID or name
        """
        return self.__destinationLakehouse
    @destinationLakehouse.setter(key='destinationLakehouse')
    def destinationLakehouse(self, val):
        self.__destinationLakehouse = val


from bifabrik.src.DataSource import DataSource
from bifabrik.cfg.JsonSourceConfiguration import JsonSourceConfiguration
from pyspark.sql.dataframe import DataFrame
from bifabrik.utils import fsUtils
import bifabrik.utils.log as lg

class JsonSource(DataSource, JsonSourceConfiguration):
    """CSV data source. Uses the spark loader in the backend.
    Supports (most of) the options from https://spark.apache.org/docs/3.1.2/api/python/reference/api/pyspark.sql.DataFrameReader.json.html

    Examples
    --------
    > import bifabrik as bif
    >
    > bif.fromJson.path('Files/JSON/ITA_TabOrder.json').multiLine(True).toTable('TabOrder1').run()
    """
    
    def __init__(self, parentPipeline):
        super().__init__(parentPipeline)
        JsonSourceConfiguration.__init__(self)
        self._path = ""

    def __str__(self):
        return f'CSV source: {self._path}'
    
    def path(self, path: str):
        """Set the path (or pattern) to the source file.
        It searches the Files/ folder in the current lakehouse
        """
        self._path = path
        return self
    
    # if key in other.propDict[key]._explicitProps:
    # df = spark.read.option("multiLine", "true").option('mode', 'PERMISSIVE').json("Files/JSON/ITA_TabZakazka.json")

    def execute(self, input) -> DataFrame:
        lgr = lg.getLogger()
        mergedConfig = self._pipeline.configuration.mergeToCopy(self)

        # find source files

        srcLh = mergedConfig.sourceStorage.sourceLakehouse
        srcWs = mergedConfig.sourceStorage.sourceWorkspace

        source_files = fsUtils.filePatternSearch(self._path, srcLh, srcWs)
        if len(source_files) == 0:
            return None
        
        separator = ', '
        lgr.info(f'Loading JSON files: [{separator.join(source_files)}]')
        
        # set spark options for all the configuration switches specified in the task's config
        readerBase = self._spark.read
        for key in mergedConfig.fileSource._explicitProps:
            #lgr.info(f'Setting {key} to {mergedConfig.fileSource._explicitProps[key]}')
            #print(f'Setting {key} to {mergedConfig.fileSource._explicitProps[key]}')
            readerBase.option(key, mergedConfig.fileSource._explicitProps[key])
        for key in mergedConfig.json._explicitProps:
            #lgr.info(f'Setting {key} to {mergedConfig.json._explicitProps[key]}')
            #print(f'Setting {key} to {mergedConfig.json._explicitProps[key]}')
            readerBase.option(key, mergedConfig.json._explicitProps[key])

        df = readerBase.json(source_files)
        
        self._result = df
        self._completed = True


#from pyspark.sql.session import SparkSession
from pyspark.sql.dataframe import DataFrame
from bifabrik.dst.TableDestination import TableDestination
from bifabrik.base.Task import Task

class DataSource(Task):
    """Generic data source. 
    Contains methods to initiate a DataDestination after you are done setting the source.
    bif.sourceInit.sourceOption1(A).srcOpt2(B).toTable("Tab").destinationOption(X)....save()
    """
    def __init__(self, parentPipeline):
        super().__init__(parentPipeline)
        

    def toDf(self) -> DataFrame:
        """Converts the data to a dataframe
        """
        pass

    def toTable(self, targetTableName: str) -> TableDestination:
        """Sets the destination table name (table in the current lakehouse)
        """
        dst = TableDestination(self._pipeline, targetTableName)
        return dst



from bifabrik.base.Task import Task
from bifabrik.cfg.CompleteConfiguration import CompleteConfiguration
import uuid
import bifabrik.utils.log as lg

class Pipeline:
    
    def __init__(self, spark, configuration: CompleteConfiguration) -> None:
        self._tasks = []
        self._id = str(uuid.uuid4())
        self.spark = spark
        self.configuration = configuration

    #t: bifabrik.base.Task.Task
    def addTask(self, t: Task):
        self._tasks.append(t)

    def getTasks(self) -> list[Task]:
        return self._tasks
    
    def getTaskResult(self, task) -> any:
        """Execute all tasks in the pipeline up to the given one and return its result
        """
        targetIdx = self._tasks.index(task)
        self._executeUpToIndex(targetIdx)
        return task.result
    
    def execute(self) -> any:
        taskCount = len(self._tasks)
        if taskCount == 0:
            return None
        
        self.clearResults()
        self.getTaskResult(self._tasks[taskCount - 1])

    def cleanup(self):
        for ix in range(0, len(self._tasks) - 1):
            self._tasks[ix].cleanup()

    def clearResults(self):
        for t in self._tasks:
            t.clearResults()
    
    def _executeUpToIndex(self, index: int):
        prevResult = None
        lgr = lg.getLogger()

        for ix in range(0, index+1):
            tsk = self._tasks[ix]
            try:
                if tsk.completed == True:
                    if tsk.error != None:
                        raise Exception(tsk.error)
                    prevResult = tsk.result
                    continue
                
                if tsk.error != None:
                    raise Exception(tsk.error)
                
                lgr.info(f'Executing {tsk}')
                tsk.execute(prevResult)
                prevResult = tsk.result
            except Exception as e:
                lgr.exception(e)
                raise e
    @property
    def id(self) -> str:
        return self._id




# from bifabrik.src.DataSource import DataSource
# from bifabrik.dst.TableDestination import TableDestination
# from pyspark.sql.session import SparkSession
# from pyspark.sql.dataframe import DataFrame
        
# class DataLoader:
#     def __init__(self, spark: SparkSession):
#         self.spark = spark
#         self.source = None
#         self.destination = None
    

# class Task:
#     _pipeline = None
#     def __init__(self, parentPipeline):
        
#         if parentPipeline != None:
#             parentPipeline.addTask(self)
        

#     def prepare():
#         pass

#     def execute():
#         pass

#     def rollback():
#         pass




"""The starting point for using the library.

    For more info see https://rjankovic.github.io/bifabrik/
    
    To create an instance, pass the :class:`SparkSession` ("spark") from your Fabric notebook.
    You can then use various methods to load data to tables to the current lakehouse.
    The notebook has to have a default lakehouse selected - it references files from the Files/
    path of the lakehouse and tables from Tables/

    Examples
    --------
    Load a CSV file from Files/Folder/orders*.csv (all files matching the pattern) to a table called OrdersTable.

    >>> import bifabrik as bif
    >>>
    >>> bif.fromCsv.path('Folder/orders*.csv').toTable('OrdersTable').run()
    
    Load the results of a SQL query to a table

    >>> bif.fromSql.query("SELECT * FROM OrdersTable LIMIT 10").toTable('TenOrders').save()
"""

from bifabrik.src.CsvSource import CsvSource
from bifabrik.src.JsonSource import JsonSource
from bifabrik.src.SqlSource import SqlSource
from bifabrik.cfg.CompleteConfiguration import CompleteConfiguration
import bifabrik.utils.log as log
from bifabrik.base.Pipeline import Pipeline
import pyspark.sql.session as pss

# read version from installed package
from importlib.metadata import version
__version__ = version("bifabrik")
__loggername__ = "bifabrik_logger"

#import bifabrik
from bifabrik.utils.fsUtils import getDefaultLakehouseAbfsPath

if getDefaultLakehouseAbfsPath() is None:
    print('bifabrik warning: the notebook is not attached to a lakehouse - some features of bifabrik will not work correctly.')


#from bifabrik.base.Task import Task

__spark = pss.SparkSession.builder.getOrCreate()
__configuration = CompleteConfiguration()
        #self.__configuration.log.loggingEnabled = False
    
def __prepPipeline() -> Pipeline:
    global __configuration
    global __spark
    log.configureLogger(__configuration.log)
    return Pipeline(__spark, __configuration)


def fromCsv(path: str = None) -> CsvSource:
    """Load data from CSV
    
    Examples
    --------
    
    >>> import bifabrik as bif
    >>>
    >>> bif.fromCsv('orders*.csv').toTable('Orders').run()
    """
    ds = CsvSource(__prepPipeline())
    ds.path(path)
    return ds


def fromJson(path: str = None) -> JsonSource:
    """Load data from JSON
    
    Examples
    --------
    
    >>> import bifabrik as bif
    >>>
    >>> bif.fromJson('invoices.json').toTable('Invoices').run()
    """
    ds = JsonSource(__prepPipeline())
    ds.path(path)
    return ds


def fromSql(query: str = None) -> SqlSource:
    """Load the result of a SQL query to a table
    
    Examples
    --------
    
    >>> import bifabrik as bif
    >>>
    >>> bif.fromSql('SELECT A, B, C FROM Table1 WHERE D = 1').toTable('Table2').run()
    """
    ds = SqlSource(__prepPipeline())
    ds.query(query)
    return ds

def __getattr__(name):
    if name == 'config':
        return __configuration
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")


from bifabrik.cfg.engine.ConfigContainer import ConfigContainer
from bifabrik.cfg.specific.SourceStorageConfiguration import SourceStorageConfiguration

class DataSourceConfigurationBase(ConfigContainer):

    def __init__(self):
        self.__sourceStorage = SourceStorageConfiguration()
        super().__init__()
        

    @property
    def sourceStorage(self) -> SourceStorageConfiguration:
        """The source lakehouse and warehouse reference used by pipelines"""
        return self.__sourceStorage


from bifabrik.cfg.engine.Configuration import Configuration
from bifabrik.cfg.engine.Configuration import CfgProperty

class MetadataStorageConfiguration(Configuration):
    """Defines the data metadata lakehouse. By default refers to the default lakehouse of the notebook.
    This is used for logs and metadata tables
    """
    def __init__(self):
        self._explicitProps = {}
        self.__metadataWorkspaceName = None
        self.__metadataLakehouseName = None

    @CfgProperty
    def metadataWorkspaceName(self) -> str:
        """The workspace where bifabrik stores metadata
        """
        return self.__metadataWorkspaceName
    @metadataWorkspaceName.setter(key='metadataWorkspaceName')
    def metadataWorkspaceName(self, val):
        self.__metadataWorkspaceName = val

    @CfgProperty
    def metadataLakehouseName(self) -> str:
        """The lakehouse where bifabrik stores metadata
        """
        return self.__metadataLakehouseName
    @metadataLakehouseName.setter(key='metadataLakehouseName')
    def metadataLakehouseName(self, val):
        self.__metadataLakehouseName = val